#!/bin/bash
# Generate binary data and corresponding labels from a given original data matrix 
# with different methods and number of bins 
#

BASE_REP=../
SCRIPTS_PATH=${BASE_REP}scripts/
MINE_SCRIPT=${SCRIPTS_PATH}greedyRedescriptions.py
ORG_REP=${BASE_REP}bio/
FILE_L=mammals_small
FILE_R=worldclim_tp_small

DATA_REP=${BASE_REP}binning/data/
TEMPLATE_CONF=${SCRIPTS_PATH}binning_template.conf
RESULTS_REP=${BASE_REP}binning/results/

DENSENUM=.densenum
DATBOOL=.datbool
NAMES=.names
BOUNDS=.bounds
SUFF_ORG=_org

MATLAB_BIN=/opt/matlab/bin/matlab
MINE_CONF=${BASE_REP}binning/binning_template.conf
TODO_FILE=${BASE_REP}binning/todo.list

mkdir -p ${DATA_REP}
mkdir -p ${RESULTS_REP}
cp ${ORG_REP}${FILE_L}${DATBOOL} ${DATA_REP}${FILE_L}${DATBOOL}
cp ${ORG_REP}${FILE_L}${NAMES} ${DATA_REP}${FILE_L}${NAMES}
cp ${ORG_REP}${FILE_R}${DENSENUM} ${DATA_REP}${FILE_R}${SUFF_ORG}${DENSENUM}
cp ${ORG_REP}${FILE_R}${NAMES} ${DATA_REP}${FILE_R}${SUFF_ORG}${NAMES}
cp $TEMPLATE_CONF $MINE_CONF
echo "0	$MINE_SCRIPT $MINE_CONF ::DATA_REP::=$DATA_REP,::RESULTS_REP::=$RESULTS_REP,::SERIES::=$SUFF_ORG,::EXT_R::=$DENSENUM" > $TODO_FILE 


for HOW in 'height' #'width' 'segments'
do
    for N in 10 20 #30 50 75 100 150
    do
       SUFF=_${N}${HOW}
       NAME_BASE=${DATA_REP}/${FILE_R}${SUFF}
       echo "Doing $SUFF"
       if [ -e ${DATA_REP}/${FILE_R}${SUFF}${BOUNDS} ]
       then
	 echo "Removing old bounds file..."
	 rm ${DATA_REP}/${FILE_R}${SUFF}${BOUNDS}
       fi
       

SCRIPT_MATLAB="
    path(path,'${SCRIPTS_PATH}');
    mat_real=load('${DATA_REP}/${FILE_R}${SUFF_ORG}${DENSENUM}')';
    [mat_bin, bounds] = bin_matrix(mat_real, ${N}, '${HOW}');
    [i,j,s] = find(mat_bin);
    i = [size(mat_bin,1); i];
    j = [size(mat_bin,2); j];
    s = [0; s~=0];
    A = [i j s];
    dlmwrite('${DATA_REP}/${FILE_R}${SUFF}${DATBOOL}', A, 'delimiter', '\t');
    for column=1:length(bounds)
        dlmwrite('${DATA_REP}/${FILE_R}${SUFF}${BOUNDS}', bounds{column}', 'delimiter', '\t','-append');    
    end
"

    echo "Computing bins..."
    echo "${SCRIPT_MATLAB}" | $MATLAB_BIN -nojvm -nosplash > /dev/null	

    echo "Translating names..."
    awk -f ${SCRIPTS_PATH}bins_names.awk sparseg=0 ${DATA_REP}/${FILE_R}${SUFF_ORG}${NAMES} ${DATA_REP}/${FILE_R}${SUFF}${BOUNDS} > ${DATA_REP}/${FILE_R}${SUFF}${NAMES}
    
    echo "0	$MINE_SCRIPT $MINE_CONF ::DATA_REP::=$DATA_REP,::RESULTS_REP::=$RESULTS_REP,::SERIES::=$SUFF,::EXT_R::=$DATBOOL" >> $TODO_FILE 

    done
done




##
##    Feb, 2011. EG. ReReMi 1.0
##
